@@ -29,10 +29,10 @@
 #include <worker.h>
 #include <main.h>
 
-const char *human_addr2(const struct sockaddr *sa, socklen_t salen,
+char *human_addr2(const struct sockaddr *sa, socklen_t salen,
 		       void *_buf, size_t buflen, unsigned full)
 {
-	const char *save_buf = _buf;
+	char *save_buf = _buf;
 	char *buf = _buf;
 	size_t l;
 
@@ -49,6 +49,14 @@ const char *human_addr2(const struct sockaddr *sa, socklen_t salen,
 	if (getnameinfo(sa, salen, buf, buflen, NULL, 0, NI_NUMERICHOST) != 0)
 		return NULL;
 
+	if (salen == sizeof(struct sockaddr_in6)) {
+		char *p = strchr(buf, '%');
+		/* remove any zone info */
+		if (p != NULL) {
+			*p = 0;
+		}
+	}
+
 	if (full == 0)
 		goto finish;
 
@@ -91,10 +99,15 @@ void __attribute__ ((format(printf, 3, 4)))
 		return;
 
 	if (priority == LOG_HTTP_DEBUG) {
-	    if (ws->config->http_debug == 0)
+	    if (ws->config->debug < DEBUG_HTTP)
                 return;
             else
                 priority = LOG_INFO;
+        } else if (priority == LOG_TRANSFER_DEBUG) {
+	    if (ws->config->debug < DEBUG_TRANSFERRED)
+                return;
+            else
+                priority = LOG_DEBUG;
         }
 
 	ip = human_addr((void*)&ws->remote_addr, ws->remote_addr_len,
@@ -105,9 +118,9 @@ void __attribute__ ((format(printf, 3, 4)))
 	va_end(args);
 
 	if (ip) {
-		syslog(priority, "%s %s", ip, buf);
+		syslog(priority, "worker: %s %s", ip, buf);
 	} else {
-		syslog(priority, "[unknown] %s", buf);
+		syslog(priority, "worker: [unknown] %s", buf);
 	}
 
 	return;
@@ -127,7 +140,12 @@ void __attribute__ ((format(printf, 4, 5)))
 		return;
 
 	if (priority == LOG_HTTP_DEBUG) {
-	    if (s->config->http_debug == 0)
+	    if (s->config->debug < DEBUG_HTTP)
+                return;
+            else
+                priority = LOG_DEBUG;
+        } else if (priority == LOG_TRANSFER_DEBUG) {
+	    if (s->config->debug < DEBUG_TRANSFERRED)
                 return;
             else
                 priority = LOG_DEBUG;
@@ -143,9 +161,9 @@ void __attribute__ ((format(printf, 4, 5)))
 	va_end(args);
 
 	if (ip) {
-		syslog(priority, "%s %s", ip, buf);
+		syslog(priority, "main: %s %s", ip, buf);
 	} else {
-		syslog(priority, "[main] %s", buf);
+		syslog(priority, "main: %s", buf);
 	}
 
 	return;
