@@ -31,7 +31,12 @@
 #include <errno.h>
 #include <sys/ioctl.h>
 #include <sys/resource.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <cloexec.h>
+#ifdef HAVE_MALLOC_TRIM
+# include <malloc.h> /* for malloc_trim() */
+#endif
 #include <script-list.h>
 
 #include <gnutls/x509.h>
@@ -46,8 +51,9 @@
 # include <systemd/sd-daemon.h>
 #endif
 #include <main.h>
+#include <main-sup-config.h>
+#include <main-ctl.h>
 #include <route-add.h>
-#include <main-auth.h>
 #include <worker.h>
 #include <cookies.h>
 #include <tun.h>
@@ -60,32 +66,18 @@ static unsigned int terminate = 0;
 static unsigned int reload_conf = 0;
 unsigned int need_maintenance = 0;
 static unsigned int need_children_cleanup = 0;
+sigset_t sig_default_set;
 
-static void ms_sleep(unsigned ms)
-{
-  struct timespec tv;
-
-  tv.tv_sec = 0;
-  tv.tv_nsec = ms * 1000 * 1000;
-
-  while(tv.tv_nsec >= 1000000000) {
-  	tv.tv_nsec -= 1000000000;
-  	tv.tv_sec++;
-  }
-  
-  nanosleep(&tv, NULL);
-}
-
-static void add_listener(struct listen_list_st *list,
+static void add_listener(void *pool, struct listen_list_st *list,
 	int fd, int family, int socktype, int protocol,
 	struct sockaddr* addr, socklen_t addr_len)
 {
 	struct listener_st *tmp;
 
-	tmp = calloc(1, sizeof(struct listener_st));
+	tmp = talloc_zero(pool, struct listener_st);
 	tmp->fd = fd;
 	tmp->family = family;
-	tmp->socktype = socktype;
+	tmp->sock_type = socktype;
 	tmp->protocol = protocol;
 	
 	tmp->addr_len = addr_len;
@@ -95,39 +87,33 @@ static void add_listener(struct listen_list_st *list,
 	list->total++;
 }
 
-static void set_udp_socket_options(int fd)
+static void set_udp_socket_options(struct cfg_st* config, int fd)
 {
 int y;
+	if (config->try_mtu) {
 #if defined(IP_DONTFRAG)
-	y = 1;
-	if (setsockopt(fd, SOL_IP, IP_DONTFRAG,
-		       (const void *) &y, sizeof(y)) < 0)
-		perror("setsockopt(IP_DF) failed");
+		y = 1;
+		if (setsockopt(fd, SOL_IP, IP_DONTFRAG,
+			       (const void *) &y, sizeof(y)) < 0)
+			perror("setsockopt(IP_DF) failed");
 #elif defined(IP_MTU_DISCOVER)
-	y = IP_PMTUDISC_DO;
-	if (setsockopt(fd, IPPROTO_IP, IP_MTU_DISCOVER,
+		y = IP_PMTUDISC_DO;
+		if (setsockopt(fd, IPPROTO_IP, IP_MTU_DISCOVER,
 		       (const void *) &y, sizeof(y)) < 0)
-		perror("setsockopt(IP_DF) failed");
+			perror("setsockopt(IP_DF) failed");
 #endif
+	}
 }
 
 static void set_common_socket_options(int fd)
 {
-int val;
-
-	val = fcntl(fd, F_GETFL, 0);
-	if ((val == -1)
-	    || (fcntl(fd, F_SETFL, val | O_NONBLOCK) < 0)) {
-		int e = errno;
-		fprintf(stderr, "fcntl() error: %s", strerror(e));
-		exit(1);
-	}
-
+	set_non_block(fd);
 	set_cloexec_flag (fd, 1);
 }
 
 static 
-int _listen_ports(struct cfg_st* config, struct addrinfo *res, struct listen_list_st *list)
+int _listen_ports(void *pool, struct cfg_st* config, 
+		struct addrinfo *res, struct listen_list_st *list)
 {
 	struct addrinfo *ptr;
 	int s, y;
@@ -174,7 +160,7 @@ int _listen_ports(struct cfg_st* config, struct addrinfo *res, struct listen_lis
 		}
 
 		if (ptr->ai_socktype == SOCK_DGRAM) {
-			set_udp_socket_options(s);
+			set_udp_socket_options(config, s);
 		}
 
 
@@ -185,19 +171,76 @@ int _listen_ports(struct cfg_st* config, struct addrinfo *res, struct listen_lis
 		}
 
 		if (ptr->ai_socktype == SOCK_STREAM) {
-			if (listen(s, 10) < 0) {
+			if (listen(s, 1024) < 0) {
 				perror("listen() failed");
+				close(s);
 				return -1;
 			}
 		}
 
 		set_common_socket_options(s);
 		
-		add_listener(list, s, ptr->ai_family, ptr->ai_socktype,
+		add_listener(pool, list, s, ptr->ai_family, ptr->ai_socktype==SOCK_STREAM?SOCK_TYPE_TCP:SOCK_TYPE_UDP,
 			ptr->ai_protocol, ptr->ai_addr, ptr->ai_addrlen);
 
 	}
-	
+
+	fflush(stderr);
+
+	return 0;
+}
+
+static 
+int _listen_unix_ports(void *pool, struct cfg_st* config, 
+		       struct listen_list_st *list)
+{
+	int s, e, ret;
+	struct sockaddr_un sa;
+
+	/* open the UNIX domain socket to accept connections */
+	if (config->unix_conn_file) {
+		memset(&sa, 0, sizeof(sa));
+		sa.sun_family = AF_UNIX;
+		snprintf(sa.sun_path, sizeof(sa.sun_path), "%s", config->unix_conn_file);
+		remove(sa.sun_path);
+
+		if (config->foreground != 0)
+			fprintf(stderr, "listening (UNIX) on %s...\n",
+				sa.sun_path);
+
+		s = socket(AF_UNIX, SOCK_STREAM, 0);
+		if (s == -1) {
+			e = errno;
+			fprintf(stderr, "could not create socket '%s': %s", sa.sun_path,
+			       strerror(e));
+			return -1;
+		}
+
+		umask(006);
+		ret = bind(s, (struct sockaddr *)&sa, SUN_LEN(&sa));
+		if (ret == -1) {
+			e = errno;
+			fprintf(stderr, "could not bind socket '%s': %s", sa.sun_path,
+			       strerror(e));
+			return -1;
+		}
+
+		ret = chown(sa.sun_path, config->uid, config->gid);
+		if (ret == -1) {
+			e = errno;
+			fprintf(stderr, "could not chown socket '%s': %s", sa.sun_path,
+			       strerror(e));
+		}
+
+		ret = listen(s, 1024);
+		if (ret == -1) {
+			e = errno;
+			fprintf(stderr, "could not listen to socket '%s': %s",
+			       sa.sun_path, strerror(e));
+			exit(1);
+		}
+		add_listener(pool, list, s, AF_UNIX, SOCK_TYPE_UNIX, 0, (struct sockaddr *)&sa, sizeof(sa));
+	}
 	fflush(stderr);
 
 	return 0;
@@ -206,7 +249,8 @@ int _listen_ports(struct cfg_st* config, struct addrinfo *res, struct listen_lis
 /* Returns 0 on success or negative value on error.
  */
 static int
-listen_ports(struct cfg_st* config, struct listen_list_st *list, const char *node)
+listen_ports(void *pool, struct cfg_st* config, 
+		struct listen_list_st *list)
 {
 	struct addrinfo hints, *res;
 	char portname[6];
@@ -246,7 +290,7 @@ listen_ports(struct cfg_st* config, struct listen_list_st *list, const char *nod
 			}
 
 			if (type == SOCK_DGRAM)
-				set_udp_socket_options(fd);
+				set_udp_socket_options(config, fd);
 
 			/* obtain socket params */
 			tmp_sock_len = sizeof(tmp_sock);
@@ -270,7 +314,7 @@ listen_ports(struct cfg_st* config, struct listen_list_st *list, const char *nod
 					config->udp_port = ntohs(((struct sockaddr_in6*)&tmp_sock)->sin6_port);
 			}
 
-			add_listener(list, fd, family, type, 0, (struct sockaddr*)&tmp_sock, tmp_sock_len);
+			add_listener(pool, list, fd, family, type==SOCK_STREAM?SOCK_TYPE_TCP:SOCK_TYPE_UDP, 0, (struct sockaddr*)&tmp_sock, tmp_sock_len);
 		}
 
 		if (list->total == 0) {
@@ -285,37 +329,44 @@ listen_ports(struct cfg_st* config, struct listen_list_st *list, const char *nod
 	}
 #endif
 
-	if (config->port == 0) {
+	if (config->port == 0 && config->unix_conn_file == NULL) {
 		fprintf(stderr, "tcp-port option is mandatory!\n");
 		return -1;
 	}
 
-	snprintf(portname, sizeof(portname), "%d", config->port);
+	if (config->port != 0) {
+		snprintf(portname, sizeof(portname), "%d", config->port);
 
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_flags = AI_PASSIVE
+		memset(&hints, 0, sizeof(hints));
+		hints.ai_socktype = SOCK_STREAM;
+		hints.ai_flags = AI_PASSIVE
 #ifdef AI_ADDRCONFIG
-	    | AI_ADDRCONFIG
+		    | AI_ADDRCONFIG
 #endif
-	    ;
+		    ;
 
-	ret = getaddrinfo(node, portname, &hints, &res);
-	if (ret != 0) {
-		fprintf(stderr, "getaddrinfo() failed: %s\n",
-			gai_strerror(ret));
-		return -1;
+		ret = getaddrinfo(config->name, portname, &hints, &res);
+		if (ret != 0) {
+			fprintf(stderr, "getaddrinfo() failed: %s\n",
+				gai_strerror(ret));
+			return -1;
+		}
+
+		ret = _listen_ports(pool, config, res, list);
+		if (ret < 0) {
+			return -1;
+		}
+
+		freeaddrinfo(res);
 	}
 
-	ret = _listen_ports(config, res, list);
+	ret = _listen_unix_ports(pool, config, list);
 	if (ret < 0) {
 		return -1;
 	}
 
-	freeaddrinfo(res);
-
 	if (list->total == 0) {
-		fprintf(stderr, "Could not listen to any TCP ports\n");
+		fprintf(stderr, "Could not listen to any TCP or UNIX ports\n");
 		exit(1);
 	}
 
@@ -330,14 +381,14 @@ listen_ports(struct cfg_st* config, struct listen_list_st *list, const char *nod
 #endif
 		    ;
 
-		ret = getaddrinfo(node, portname, &hints, &res);
+		ret = getaddrinfo(config->name, portname, &hints, &res);
 		if (ret != 0) {
 			fprintf(stderr, "getaddrinfo() failed: %s\n",
 				gai_strerror(ret));
 			return -1;
 		}
 
-		ret = _listen_ports(config, res, list);
+		ret = _listen_ports(pool, config, res, list);
 		if (ret < 0) {
 			return -1;
 		}
@@ -361,7 +412,7 @@ int s, y, e;
 	close(l->fd);
 	l->fd = -1;
 
-	s = socket(l->family, l->socktype, l->protocol);
+	s = socket(l->family, SOCK_DGRAM, l->protocol);
 	if (s < 0) {
 		perror("socket() failed");
 		return -1;
@@ -424,9 +475,11 @@ struct script_wait_st *stmp = NULL, *spos;
 				mslog(s, stmp->proc, LOG_DEBUG, "%s-script exit status: %u", stmp->up?"connect":"disconnect", estatus);
 				list_del(&stmp->list);
 				ret = handle_script_exit(s, stmp->proc, estatus);
-				if (ret < 0)
+				if (ret < 0) {
 					remove_proc(s, stmp->proc, 1);
-				free(stmp);
+				} else {
+					talloc_free(stmp);
+				}
 				break;
 			}
 		}
@@ -544,49 +597,47 @@ void clear_lists(main_server_st *s)
 	struct listener_st *ltmp = NULL, *lpos;
 	struct proc_st *ctmp = NULL, *cpos;
 	struct script_wait_st *script_tmp = NULL, *script_pos;
-	struct banned_st *btmp = NULL, *bpos;
 
 	list_for_each_safe(&s->listen_list.head, ltmp, lpos, list) {
 		close(ltmp->fd);
 		list_del(&ltmp->list);
-		free(ltmp);
+		talloc_free(ltmp);
 		s->listen_list.total--;
 	}
 
 	list_for_each_safe(&s->proc_list.head, ctmp, cpos, list) {
 		if (ctmp->fd >= 0)
 			close(ctmp->fd);
+		if (ctmp->tun_lease.fd >= 0)
+			close(ctmp->tun_lease.fd);
+		if (ctmp->cookie_ptr)
+			ctmp->cookie_ptr->proc = NULL;
 		list_del(&ctmp->list);
-		memset(ctmp, 0, sizeof(*ctmp));
-		free(ctmp);
+		safe_memset(ctmp, 0, sizeof(*ctmp));
+		talloc_free(ctmp);
 		s->proc_list.total--;
 	}
 
-	list_for_each_safe(&s->ban_list.head, btmp, bpos, list) {
-		list_del(&btmp->list);
-		free(btmp);
-	}
-
 	list_for_each_safe(&s->script_list.head, script_tmp, script_pos, list) {
 		list_del(&script_tmp->list);
-		free(script_tmp);
+		talloc_free(script_tmp);
 	}
 
-	tls_cache_deinit(s->tls_db);
+	tls_cache_deinit(&s->tls_db);
 	ip_lease_deinit(&s->ip_leases);
 	ctl_handler_deinit(s);
+	cookie_db_deinit(&s->cookies);
 }
 
 static void kill_children(main_server_st* s)
 {
-	struct proc_st *ctmp = NULL;
+	struct proc_st *ctmp = NULL, *cpos;
 
 	/* kill the security module server */
 	kill(s->sec_mod_pid, SIGTERM);
-	list_for_each(&s->proc_list.head, ctmp, list) {
+	list_for_each_safe(&s->proc_list.head, ctmp, cpos, list) {
 		if (ctmp->pid != -1) {
-			kill(ctmp->pid, SIGTERM);
-			user_disconnected(s, ctmp);
+			remove_proc(s, ctmp, 1);
 		}
 	}
 }
@@ -613,13 +664,15 @@ static int forward_udp_to_owner(main_server_st* s, struct listener_st *listener)
 {
 int ret, e;
 struct sockaddr_storage cli_addr;
-struct proc_st *ctmp = NULL;
+struct proc_st *ctmp = NULL, *proc_to_send = NULL;
 socklen_t cli_addr_size;
 uint8_t buffer[1024];
-uint8_t  *session_id;
-int session_id_size;
+char tbuf[64];
+uint8_t  *session_id = NULL;
+int session_id_size = 0;
 ssize_t buffer_size;
 int connected = 0;
+int match_ip_only = 0, matching_ips;
 time_t now;
 
 	/* first receive from the correct client and connect socket */
@@ -633,73 +686,119 @@ time_t now;
 	buffer_size = ret;
 	
 	/* obtain the session id */
-	if (buffer_size < RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS+GNUTLS_MAX_SESSION_ID+2)
+	if (buffer_size < RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS+GNUTLS_MAX_SESSION_ID+2) {
+		mslog(s, NULL, LOG_INFO, "%s: too short packet",
+		      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
 		goto fail;
+	}
 
 	/* check version */
-	mslog(s, NULL, LOG_DEBUG, "DTLS record version: %u.%u", (unsigned int)buffer[1], (unsigned int)buffer[2]);
-	mslog(s, NULL, LOG_DEBUG, "DTLS hello version: %u.%u", (unsigned int)buffer[RECORD_PAYLOAD_POS], (unsigned int)buffer[RECORD_PAYLOAD_POS+1]);
+	mslog(s, NULL, LOG_DEBUG, "new DTLS session from %s (record v%u.%u, hello v%u.%u)", 
+		human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)),
+		(unsigned int)buffer[1], (unsigned int)buffer[2],
+		(unsigned int)buffer[RECORD_PAYLOAD_POS], (unsigned int)buffer[RECORD_PAYLOAD_POS+1]);
+
 	if (buffer[1] != 254 && (buffer[1] != 1 && buffer[2] != 0) &&
 		buffer[RECORD_PAYLOAD_POS] != 254 && (buffer[RECORD_PAYLOAD_POS] != 0 && buffer[RECORD_PAYLOAD_POS+1] != 0)) {
-		mslog(s, NULL, LOG_INFO, "unknown DTLS version: %u.%u", (unsigned)buffer[1], (unsigned)buffer[2]);
+		mslog(s, NULL, LOG_INFO, "%s: unknown DTLS version: %u.%u", 
+		      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)),
+		      (unsigned)buffer[1], (unsigned)buffer[2]);
 		goto fail;
 	}
 	if (buffer[0] != 22) {
-		mslog(s, NULL, LOG_INFO, "unexpected DTLS content type: %u", (unsigned int)buffer[0]);
-		goto fail;
+		mslog(s, NULL, LOG_INFO, "%s: unexpected DTLS content type: %u",
+		      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)),
+		      (unsigned int)buffer[0]);
+		/* Here we received a non-client hello packet. It may be that
+		 * the client's NAT changed its UDP source port and the previous
+		 * connection is invalidated. Try to see if we can simply match
+		 * the IP address and forward the socket.
+		 */
+		match_ip_only = 1;
+	} else {
+		/* read session_id */
+		session_id_size = buffer[RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS];
+		session_id = &buffer[RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS+1];
 	}
 
-	/* read session_id */
-	session_id_size = buffer[RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS];
-	session_id = &buffer[RECORD_PAYLOAD_POS+HANDSHAKE_SESSION_ID_POS+1];
-
 	/* search for the IP and the session ID in all procs */
 	now = time(0);
+	matching_ips = 0;
 
 	list_for_each(&s->proc_list.head, ctmp, list) {
-		if (session_id_size == ctmp->dtls_session_id_size &&
-			memcmp(session_id, ctmp->dtls_session_id, session_id_size) == 0 &&
-			(now - ctmp->udp_fd_receive_time > UDP_FD_RESEND_TIME)) {
-			UdpFdMsg msg = UDP_FD_MSG__INIT;
+		if (match_ip_only == 0 && session_id_size == ctmp->dtls_session_id_size &&
+			memcmp(session_id, ctmp->dtls_session_id, session_id_size) == 0) {
 
-			ret = connect(listener->fd, (void*)&cli_addr, cli_addr_size);
-			if (ret == -1) {
-				e = errno;
-				mslog(s, ctmp, LOG_ERR, "connect UDP socket: %s", strerror(e));
-				return -1;
-			}
-
-			ret = send_socket_msg_to_worker(s, ctmp, CMD_UDP_FD,
-				listener->fd,
-				&msg, 
-				(pack_size_func)udp_fd_msg__get_packed_size,
-				(pack_func)udp_fd_msg__pack);
-			if (ret < 0) {
-				mslog(s, ctmp, LOG_ERR, "error passing UDP socket");
-				return -1;
-			}
-			mslog(s, ctmp, LOG_DEBUG, "passed UDP socket");
-			ctmp->udp_fd_receive_time = now;
-			connected = 1;
-			
-			reopen_udp_port(listener);
+			proc_to_send = ctmp;
 			break;
+		} else if (match_ip_only != 0 && cli_addr_size == ctmp->remote_addr_len &&
+			memcmp(SA_IN_P_GENERIC(&cli_addr, cli_addr_size), 
+				SA_IN_P_GENERIC(&ctmp->remote_addr, ctmp->remote_addr_len),
+				SA_IN_SIZE(ctmp->remote_addr_len)) == 0) {
+			matching_ips++;
+			proc_to_send = ctmp;
 		}
 	}
 
+	if (proc_to_send != 0) {
+		UdpFdMsg msg = UDP_FD_MSG__INIT;
+
+		if (matching_ips > 1) {
+			mslog(s, proc_to_send, LOG_INFO, "cannot associate with a client; more than a single clients from %s",
+			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
+			goto fail;
+		}
+
+		if (now - proc_to_send->udp_fd_receive_time <= UDP_FD_RESEND_TIME) {
+			mslog(s, proc_to_send, LOG_INFO, "received UDP connection too soon from %s",
+			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
+			goto fail;
+		}
+
+		ret = connect(listener->fd, (void*)&cli_addr, cli_addr_size);
+		if (ret == -1) {
+			e = errno;
+			mslog(s, proc_to_send, LOG_ERR, "connect UDP socket from %s: %s",
+			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)),
+			      strerror(e));
+			goto fail;
+		}
+
+		if (match_ip_only != 0) {
+			msg.hello = 0;
+		}
+
+		ret = send_socket_msg_to_worker(s, proc_to_send, CMD_UDP_FD,
+			listener->fd,
+			&msg, 
+			(pack_size_func)udp_fd_msg__get_packed_size,
+			(pack_func)udp_fd_msg__pack);
+		if (ret < 0) {
+			mslog(s, proc_to_send, LOG_ERR, "error passing UDP socket from %s",
+			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
+			goto fail;
+		}
+		mslog(s, proc_to_send, LOG_DEBUG, "passed UDP socket from %s",
+		      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
+		proc_to_send->udp_fd_receive_time = now;
+		connected = 1;
+
+		reopen_udp_port(listener);
+	}
+
 fail:
 	if (connected == 0) {
-		/* received packet from unknown host */
+		/* received packet from unknown host. Ignore it. */
 		recv(listener->fd, buffer, buffer_size, 0);
 
 		return -1;
 	}
-	
+
 	return 0;
 
 }
 
-#define MAINTAINANCE_TIME(s) (MIN((300 + MAX_ZOMBIE_SECS), ((s)->config->cookie_validity + 300)))
+#define MAINTAINANCE_TIME(s) (900)
 
 static void check_other_work(main_server_st *s)
 {
@@ -707,8 +806,10 @@ unsigned total = 10;
 
 	if (reload_conf != 0) {
 		mslog(s, NULL, LOG_INFO, "reloading configuration");
-		reload_cfg_file(s->config);
+		reload_cfg_file(s->main_pool, s->config);
+		tls_reload_crl(s, s->creds);
 		reload_conf = 0;
+		kill(s->sec_mod_pid, SIGHUP);
 	}
 
 	if (need_children_cleanup != 0) {
@@ -716,25 +817,29 @@ unsigned total = 10;
 	}
 
 	if (terminate != 0) {
-		mslog(s, NULL, LOG_DEBUG, "termination request received; waiting for children to die");
+		mslog(s, NULL, LOG_INFO, "termination request received; waiting for children to die");
 		kill_children(s);
-		while (waitpid(-1, NULL, WNOHANG) == 0) {
+		remove(s->full_socket_file);
+		remove(s->config->occtl_socket_file);
+		remove_pid_file();
+
+		while (waitpid(-1, NULL, WNOHANG) >= 0) {
 			if (total == 0) {
-				mslog(s, NULL, LOG_DEBUG, "not everyone died; forcing kill");
+				mslog(s, NULL, LOG_INFO, "not everyone died; forcing kill");
 				kill(0, SIGKILL);
 			}
 			ms_sleep(500);
 			total--;
 		}
-		remove(s->socket_file);
-		remove_pid_file();
 
 		/* try to clean-up everything allocated to ease checks 
 		 * for memory leaks.
 		 */
 		clear_lists(s);
-		tls_global_deinit(s);
+		tls_global_deinit(s->creds);
 		clear_cfg_file(s->config);
+		talloc_free(s->config);
+		talloc_free(s->main_pool);
 		closelog();
 		exit(0);
 	}
@@ -742,10 +847,9 @@ unsigned total = 10;
 	/* Check if we need to expire any cookies */
 	if (need_maintenance != 0) {
 		need_maintenance = 0;
-		mslog(s, NULL, LOG_DEBUG, "Performing maintenance");
+		mslog(s, NULL, LOG_DEBUG, "performing maintenance");
 		expire_tls_sessions(s);
-		expire_zombies(s);
-		expire_banned(s);
+		expire_cookies(&s->cookies);
 		alarm(MAINTAINANCE_TIME(s));
 	}
 }
@@ -754,14 +858,14 @@ unsigned total = 10;
 static int check_tcp_wrapper(int fd)
 {
 	struct request_info req;
-	
+
 	if (request_init(&req, RQ_FILE, fd, RQ_DAEMON, PACKAGE_NAME, 0) == NULL)
 		return -1;
-	
+
 	sock_host(&req);
 	if (hosts_access(&req) == 0)
 		return -1;
-		
+
 	return 0;
 }
 #else
@@ -773,7 +877,6 @@ int main(int argc, char** argv)
 	int fd, pid, e;
 	struct listener_st *ltmp = NULL;
 	struct proc_st *ctmp = NULL, *cpos;
-	struct ctl_handler_st* ctl_tmp = NULL, *ctl_pos;
 	fd_set rd_set, wr_set;
 	int n = 0, ret, flags;
 #ifdef HAVE_PSELECT
@@ -782,20 +885,42 @@ int main(int argc, char** argv)
 	struct timeval ts;
 #endif
 	int cmd_fd[2];
-	struct worker_st ws;
-	struct cfg_st config;
+	struct worker_st *ws;
+	void *worker_pool;
+	void *main_pool;
 	unsigned set;
-	main_server_st s;
+	main_server_st *s;
 	sigset_t emptyset, blockset;
+	/* tls credentials */
+	struct tls_st creds;
 
-	memset(&s, 0, sizeof(s));
+#ifdef DEBUG_LEAKS
+	talloc_enable_leak_report_full();
+#endif
 
-	list_head_init(&s.proc_list.head);
-	list_head_init(&s.ban_list.head);
-	list_head_init(&s.script_list.head);
-	list_head_init(&s.ctl_list.head);
-	tls_cache_init(&s.tls_db);
-	ip_lease_init(&s.ip_leases);
+	memset(&creds, 0, sizeof(creds));
+
+	/* main pool */
+	main_pool = talloc_init("main");
+	if (main_pool == NULL) {
+		fprintf(stderr, "talloc init error\n");
+		exit(1);
+	}
+
+	s = talloc_zero(main_pool, main_server_st);
+	if (s == NULL) {
+		fprintf(stderr, "memory error\n");
+		exit(1);
+	}
+	s->main_pool = main_pool;
+	s->creds = &creds;
+	s->start_time = time(0);
+
+	list_head_init(&s->proc_list.head);
+	list_head_init(&s->script_list.head);
+	tls_cache_init(s, &s->tls_db);
+	cookie_db_init(s, &s->cookies);
+	ip_lease_init(&s->ip_leases);
 
 	sigemptyset(&blockset);
 	sigemptyset(&emptyset);
@@ -813,16 +938,18 @@ int main(int argc, char** argv)
 	ocsignal(SIGALRM, handle_alarm);
 
 	/* Initialize GnuTLS */
-	tls_global_init(&s);
-	
-	ret = gnutls_rnd(GNUTLS_RND_RANDOM, s.cookie_key, sizeof(s.cookie_key));
+	tls_global_init(&creds);
+
+	/* this is the key used to sign and verify cookies. It is used
+	 * by sec-mod (for signing) and main (for verification). */
+	ret = gnutls_rnd(GNUTLS_RND_RANDOM, s->cookie_key, sizeof(s->cookie_key));
 	if (ret < 0) {
 		fprintf(stderr, "Error in cookie key generation\n");
 		exit(1);
 	}
 
 	/* load configuration */
-	ret = cmd_parser(argc, argv, &config);
+	ret = cmd_parser(main_pool, argc, argv, &s->config);
 	if (ret < 0) {
 		fprintf(stderr, "Error in arguments\n");
 		exit(1);
@@ -835,12 +962,8 @@ int main(int argc, char** argv)
 		exit(1);
 	}
 
-	s.config = &config;
-
-	main_auth_init(&s);
-
 	/* Listen to network ports */
-	ret = listen_ports(&config, &s.listen_list, config.name);
+	ret = listen_ports(s, s->config, &s->listen_list);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot listen to specified ports\n");
 		exit(1);
@@ -848,7 +971,7 @@ int main(int argc, char** argv)
 
 	flags = LOG_PID|LOG_NDELAY;
 #ifdef LOG_PERROR
-	if (config.debug != 0 || config.http_debug != 0 || config.tls_debug != 0)
+	if (s->config->debug != 0)
 		flags |= LOG_PERROR;
 #endif
 	openlog("ocserv", flags, LOG_DAEMON);
@@ -858,7 +981,7 @@ int main(int argc, char** argv)
 	deny_severity = LOG_DAEMON|LOG_WARNING;
 #endif	
 
-	if (config.foreground == 0) {
+	if (s->config->foreground == 0) {
 		if (daemon(0, 0) == -1) {
 			e = errno;
 			fprintf(stderr, "daemon failed: %s\n", strerror(e));
@@ -868,52 +991,70 @@ int main(int argc, char** argv)
 
 	write_pid_file();
 	
-	run_sec_mod(&s);
+	run_sec_mod(s);
 
-	/* Initialize certificates */
-	tls_global_init_certs(&s);
+	sup_config_init(s);
 
-	mslog(&s, NULL, LOG_INFO, "initialized %s", PACKAGE_STRING);
-
-	ret = ctl_handler_init(&s);
+	ret = ctl_handler_init(s);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot create command handler\n");
 		exit(1);
 	}
 
-	sigprocmask(SIG_BLOCK, &blockset, NULL);
-	alarm(MAINTAINANCE_TIME(&s));
+	mslog(s, NULL, LOG_INFO, "initialized %s", PACKAGE_STRING);
+
+	/* chdir to our chroot directory, to allow opening the sec-mod
+	 * socket if necessary. */
+	if (s->config->chroot_dir)
+		chdir(s->config->chroot_dir);
+	ms_sleep(100); /* give some time for sec-mod to initialize */
+
+	/* Initialize certificates */
+	tls_load_certs(s, &creds);
+
+	s->secmod_addr.sun_family = AF_UNIX;
+	snprintf(s->secmod_addr.sun_path, sizeof(s->secmod_addr.sun_path), "%s", s->socket_file);
+	s->secmod_addr_len = SUN_LEN(&s->secmod_addr);
+
+	/* initialize memory for worker process */
+	worker_pool = talloc_named(main_pool, 0, "worker");
+	if (worker_pool == NULL) {
+		fprintf(stderr, "talloc init error\n");
+		exit(1);
+	}
+
+	ws = talloc_zero(worker_pool, struct worker_st);
+	if (ws == NULL) {
+		fprintf(stderr, "memory error\n");
+		exit(1);
+	}
+
+	sigprocmask(SIG_BLOCK, &blockset, &sig_default_set);
+	alarm(MAINTAINANCE_TIME(s));
 
 	for (;;) {
-		check_other_work(&s);
+		check_other_work(s);
 
 		/* initialize select */
 		FD_ZERO(&rd_set);
 		FD_ZERO(&wr_set);
 
-		list_for_each(&s.listen_list.head, ltmp, list) {
+		list_for_each(&s->listen_list.head, ltmp, list) {
 			if (ltmp->fd == -1) continue;
 
 			FD_SET(ltmp->fd, &rd_set);
 			n = MAX(n, ltmp->fd);
 		}
 
-		list_for_each(&s.proc_list.head, ctmp, list) {
+		list_for_each(&s->proc_list.head, ctmp, list) {
 			if (ctmp->fd > 0) {
 				FD_SET(ctmp->fd, &rd_set);
 				n = MAX(n, ctmp->fd);
 			}
 		}
 
-		list_for_each(&s.ctl_list.head, ctl_tmp, list) {
-			if (ctl_tmp->enabled) {
-				if (ctl_tmp->type == CTL_READ)
-					FD_SET(ctl_tmp->fd, &rd_set);
-				else
-					FD_SET(ctl_tmp->fd, &wr_set);
-				n = MAX(n, ctl_tmp->fd);
-			}
-		}
+		ret = ctl_handler_set_fds(s, &rd_set, &wr_set);
+		n = MAX(n, ret);
 
 #ifdef HAVE_PSELECT
 		ts.tv_nsec = 0;
@@ -931,52 +1072,43 @@ int main(int argc, char** argv)
 
 		if (ret < 0) {
 			e = errno;
-			mslog(&s, NULL, LOG_ERR, "Error in pselect(): %s",
+			mslog(s, NULL, LOG_ERR, "Error in pselect(): %s",
 			       strerror(e));
 			exit(1);
 		}
 
 		/* Check for new connections to accept */
-		list_for_each(&s.listen_list.head, ltmp, list) {
+		list_for_each(&s->listen_list.head, ltmp, list) {
 			set = FD_ISSET(ltmp->fd, &rd_set);
-			if (set && ltmp->socktype == SOCK_STREAM) {
+			if (set && (ltmp->sock_type == SOCK_TYPE_TCP || ltmp->sock_type == SOCK_TYPE_UNIX)) {
 				/* connection on TCP port */
-				memset(&ws, 0, sizeof(ws));
+				int stype = ltmp->sock_type;
 
-				ws.remote_addr_len = sizeof(ws.remote_addr);
-				fd = accept(ltmp->fd, (void*)&ws.remote_addr, &ws.remote_addr_len);
+				ws->remote_addr_len = sizeof(ws->remote_addr);
+				fd = accept(ltmp->fd, (void*)&ws->remote_addr, &ws->remote_addr_len);
 				if (fd < 0) {
-					mslog(&s, NULL, LOG_ERR,
+					mslog(s, NULL, LOG_ERR,
 					       "Error in accept(): %s", strerror(errno));
 					continue;
 				}
 				set_cloexec_flag (fd, 1);
 
-				/* Check if the client is on the banned list */
-				ret = check_if_banned(&s, &ws.remote_addr, ws.remote_addr_len);
-				if (ret < 0) {
-					/* banned */
+				if (s->config->max_clients > 0 && s->active_clients >= s->config->max_clients) {
 					close(fd);
-					mslog(&s, NULL, LOG_INFO, "dropping client connection due to a previous failed authentication attempt");
-					break;
-				}
-
-				if (config.max_clients > 0 && s.active_clients >= config.max_clients) {
-					close(fd);
-					mslog(&s, NULL, LOG_INFO, "Reached maximum client limit (active: %u)", s.active_clients);
+					mslog(s, NULL, LOG_INFO, "Reached maximum client limit (active: %u)", s->active_clients);
 					break;
 				}
 
 				if (check_tcp_wrapper(fd) < 0) {
 					close(fd);
-					mslog(&s, NULL, LOG_INFO, "TCP wrappers rejected the connection (see /etc/hosts.[allow|deny])");
+					mslog(s, NULL, LOG_INFO, "TCP wrappers rejected the connection (see /etc/hosts->[allow|deny])");
 					break;
 				}
 
 				/* Create a command socket */
 				ret = socketpair(AF_UNIX, SOCK_STREAM, 0, cmd_fd);
 				if (ret < 0) {
-					mslog(&s, NULL, LOG_ERR, "Error creating command socket");
+					mslog(s, NULL, LOG_ERR, "Error creating command socket");
 					close(fd);
 					break;
 				}
@@ -986,92 +1118,92 @@ int main(int argc, char** argv)
 					/* close any open descriptors, and erase
 					 * sensitive data before running the worker
 					 */
+					sigprocmask(SIG_SETMASK, &sig_default_set, NULL);
 					close(cmd_fd[0]);
-					clear_lists(&s);
+					clear_lists(s);
+
+					/* clear the cookie key */
+					safe_memset(s->cookie_key, 0, sizeof(s->cookie_key));
 
 					setproctitle(PACKAGE_NAME"-worker");
 					kill_on_parent_kill(SIGTERM);
-					
-					ws.config = &config;
-					ws.cmd_fd = cmd_fd[1];
-					ws.tun_fd = -1;
-					ws.udp_fd = -1;
-					ws.conn_fd = fd;
-					ws.creds = &s.creds;
+
+					/* write sec-mod's address */
+					memcpy(&ws->secmod_addr, &s->secmod_addr, s->secmod_addr_len);
+					ws->secmod_addr_len = s->secmod_addr_len;
+
+					ws->main_pool = main_pool;
+					ws->config = s->config;
+					ws->cmd_fd = cmd_fd[1];
+					ws->tun_fd = -1;
+					ws->udp_fd = -1;
+					ws->conn_fd = fd;
+					ws->conn_type = stype;
+					ws->creds = &creds;
 
 					/* Drop privileges after this point */
-					sigprocmask(SIG_UNBLOCK, &blockset, NULL);
-					drop_privileges(&s);
+					drop_privileges(s);
 
-					vpn_server(&ws);
+					/* creds and config are not allocated
+					 * under s.
+					 */
+					talloc_free(s);
+#ifdef HAVE_MALLOC_TRIM
+					/* try to return all the pages we've freed to
+					 * the operating system, to prevent the child from
+					 * accessing them. That's totally unreliable, so
+					 * sensitive data have to be overwritten anyway. */
+					malloc_trim(0);
+#endif
+					vpn_server(ws);
 					exit(0);
 				} else if (pid == -1) {
 fork_failed:
 					close(cmd_fd[0]);
 				} else { /* parent */
 					/* add_proc */
-					ctmp = calloc(1, sizeof(struct proc_st));
+					ctmp = new_proc(s, pid, cmd_fd[0], 
+							&ws->remote_addr, ws->remote_addr_len,
+							ws->sid, sizeof(ws->sid));
 					if (ctmp == NULL) {
 						kill(pid, SIGTERM);
 						goto fork_failed;
 					}
-					memcpy(&ctmp->remote_addr, &ws.remote_addr, ws.remote_addr_len);
-					ctmp->remote_addr_len = ws.remote_addr_len;
 
-					ctmp->pid = pid;
-					ctmp->conn_time = time(0);
-					ctmp->fd = cmd_fd[0];
-					set_cloexec_flag (cmd_fd[0], 1);
-
-					list_add(&s.proc_list.head, &(ctmp->list));
-
-					put_into_cgroup(&s, s.config->cgroup, pid);
-
-					s.active_clients++;
 				}
 				close(cmd_fd[1]);
 				close(fd);
 
-				if (config.rate_limit_ms > 0)
-					ms_sleep(config.rate_limit_ms);
-			} else if (set && ltmp->socktype == SOCK_DGRAM) {
+				if (s->config->rate_limit_ms > 0)
+					ms_sleep(s->config->rate_limit_ms);
+			} else if (set && ltmp->sock_type == SOCK_TYPE_UDP) {
 				/* connection on UDP port */
-				ret = forward_udp_to_owner(&s, ltmp);
+				ret = forward_udp_to_owner(s, ltmp);
 				if (ret < 0) {
-					mslog(&s, NULL, LOG_INFO, "could not determine the owner of received UDP packet");
+					mslog(s, NULL, LOG_INFO, "could not determine the owner of received UDP packet");
 				}
 
-				if (config.rate_limit_ms > 0)
-					ms_sleep(config.rate_limit_ms);
+				if (s->config->rate_limit_ms > 0)
+					ms_sleep(s->config->rate_limit_ms);
 			}
 		}
 
 		/* Check for any pending commands */
-		list_for_each_safe(&s.proc_list.head, ctmp, cpos, list) {
+		list_for_each_safe(&s->proc_list.head, ctmp, cpos, list) {
 			if (ctmp->fd >= 0 && FD_ISSET(ctmp->fd, &rd_set)) {
-				ret = handle_commands(&s, ctmp);
-				if (ret == ERR_WORKER_TERMINATED && ctmp->status == PS_AUTH_INIT &&
-					s.config->cisco_client_compat != 0) {
-					proc_to_zombie(&s, ctmp);
-				} else if (ret < 0) {
-					remove_proc(&s, ctmp, (ret!=ERR_WORKER_TERMINATED)?1:0);
+				ret = handle_commands(s, ctmp);
+				if (ret < 0) {
+					remove_proc(s, ctmp, (ret!=ERR_WORKER_TERMINATED)?1:0);
 				}
 			}
 		}
 
 		/* Check for pending control commands */
-		list_for_each_safe(&s.ctl_list.head, ctl_tmp, ctl_pos, list) {
-			if (ctl_tmp->enabled == 0)
-				continue;
+		ctl_handler_run_pending(s, &rd_set, &wr_set);
 
-			if (ctl_tmp->type == CTL_READ) {
-				if (FD_ISSET(ctl_tmp->fd, &rd_set))
-					ctl_handle_commands(&s, ctl_tmp);
-			} else {
-				if (FD_ISSET(ctl_tmp->fd, &wr_set))
-					ctl_handle_commands(&s, ctl_tmp);
-			}
-		}
+#ifdef DEBUG_LEAKS
+		talloc_report_full(s, stderr);
+#endif
 	}
 
 	return 0;
