@@ -28,11 +28,20 @@
 #include <vpn.h>
 #include <tlslib.h>
 #include "ipc.pb-c.h"
-#include <cookies.h>
 #include <common.h>
+#include <sys/un.h>
+#include <sys/uio.h>
 
-int cmd_parser (int argc, char **argv, struct cfg_st* config);
-void reload_cfg_file(struct cfg_st* config);
+#ifdef __FreeBSD__
+# include <limits.h>
+# define SOL_IP IPPROTO_IP
+#endif
+
+#define COOKIE_KEY_SIZE 16
+
+extern sigset_t sig_default_set;
+int cmd_parser (void *pool, int argc, char **argv, struct cfg_st** config);
+void reload_cfg_file(void *pool, struct cfg_st* config);
 void clear_cfg_file(struct cfg_st* config);
 void write_pid_file(void);
 void remove_pid_file(void);
@@ -43,7 +52,7 @@ extern unsigned int need_maintenance;
 struct listener_st {
 	struct list_node list;
 	int fd;
-	int socktype;
+	sock_type_t sock_type;
 
 	struct sockaddr_storage addr; /* local socket address */
 	socklen_t addr_len;
@@ -68,13 +77,25 @@ enum {
 	PS_AUTH_INACTIVE, /* no comm with worker */
 	PS_AUTH_FAILED, /* no tried authenticated but failed */
 	PS_AUTH_INIT, /* worker has sent an auth init msg */
-	PS_AUTH_ZOMBIE, /* in INIT state but worker has disconnected! - only present when cisco-client-compat is set */
 	PS_AUTH_COMPLETED, /* successful authentication */
 };
 
+#define COOKIE_HASH_SIZE 20
+#define COOKIE_HASH GNUTLS_DIG_SHA1
+
+typedef struct cookie_entry_st {
+	struct proc_st *proc; /* may be null, otherwise the proc that uses that cookie */
+	time_t expiration; /* -1 or the time it should expire */
+
+	/* We store the hash of the cookie that is associated with a particular session.
+	 * The reason is to avoid a memory leak to an unprivileged process to expose
+	 * data that can be used to authenticate as another use */
+	uint8_t cookie_hash[COOKIE_HASH_SIZE];
+} cookie_st;
+
 /* Each worker process maps to a unique proc_st structure.
  */
-struct proc_st {
+typedef struct proc_st {
 	struct list_node list;
 	int fd; /* the command file descriptor */
 	pid_t pid;
@@ -86,17 +107,14 @@ struct proc_st {
 	struct tun_lease_st tun_lease;
 	struct ip_lease_st *ipv4;
 	struct ip_lease_st *ipv6;
+	unsigned leases_in_use; /* someone else got our IP leases */
 
 	struct sockaddr_storage remote_addr; /* peer address */
 	socklen_t remote_addr_len;
 
-	/* A unique session identifier used to distinguish sessions
-	 * prior to authentication. It is sent as cookie to the client
-	 * who re-uses it when it performs authentication in multiple
-	 * sessions.
-	 */
+	/* The SID present in the cookie. Used for session control only */
 	uint8_t sid[SID_SIZE];
-	unsigned sid_size; /* acts as a flag if sid is set */
+	unsigned active_sid;
 
 	/* The DTLS session ID associated with the TLS session 
 	 * it is either generated or restored from a cookie.
@@ -108,7 +126,6 @@ struct proc_st {
 	char username[MAX_USERNAME_SIZE]; /* the owner */
 	char groupname[MAX_GROUPNAME_SIZE]; /* the owner's group */
 	char hostname[MAX_HOSTNAME_SIZE]; /* the requested hostname */
-	uint8_t cookie[COOKIE_SIZE]; /* the cookie associated with the session */
 
 	/* the following are copied here from the worker process for reporting
 	 * purposes (from main-ctl-handler). */
@@ -116,20 +133,26 @@ struct proc_st {
 	char tls_ciphersuite[MAX_CIPHERSUITE_NAME];
 	char dtls_ciphersuite[MAX_DTLS_CIPHERSUITE_NAME];
 
+	/* pointer to the cookie used by this session */
+	struct cookie_entry_st *cookie_ptr;
+
 	/* if the session is initiated by a cookie the following two are set
 	 * and are considered when generating an IP address. That is used to
 	 * generate the same address as previously allocated.
 	 */
-	uint8_t seeds_are_set; /* non zero if the following two elements are set */
 	uint8_t ipv4_seed[4];
 
-	void * auth_ctx; /* the context of authentication */
 	unsigned status; /* PS_AUTH_ */
-	unsigned auth_reqs; /* the number of requests received */
+	unsigned resume_reqs; /* the number of requests received */
+
+	/* these are filled in after the worker process dies, using the
+	 * Cli stats message. */
+	uint64_t bytes_in;
+	uint64_t bytes_out;
 	
 	unsigned applied_iroutes; /* whether the iroutes in the config have been successfully applied */
 	struct group_cfg_st config; /* custom user/group config */
-};
+} proc_st;
 
 struct ip_lease_db_st {
 	struct htable ht;
@@ -151,51 +174,45 @@ struct banned_st {
 	socklen_t addr_len;
 };
 
-struct ban_list_st {
-	struct list_head head;
-};
-
-#define CTL_READ 1
-#define CTL_WRITE 2
-
-struct ctl_handler_st {
-	struct list_node list;
-	int fd;
-	unsigned type; /* CTL_READ/WRITE */
-	unsigned enabled;
-	void* watch;
-};
-
-struct ctl_list_st {
-	struct list_head head;
+struct cookie_entry_db_st {
+	struct htable *db;
+	unsigned total;
 };
 
 typedef struct main_server_st {
 	struct cfg_st *config;
 	
 	struct ip_lease_db_st ip_leases;
+	struct cookie_entry_db_st cookies;
 
-	hash_db_st *tls_db;
+	tls_sess_db_st tls_db;
+	tls_st *creds;
 	
-	uint8_t cookie_key[16];
-
-	/* tls credentials */
-	struct tls_st creds;
+	uint8_t cookie_key[COOKIE_KEY_SIZE];
 
 	struct listen_list_st listen_list;
 	struct proc_list_st proc_list;
 	struct script_list_st script_list;
-	struct ban_list_st ban_list;
 	
 	char socket_file[_POSIX_PATH_MAX];
+	char full_socket_file[_POSIX_PATH_MAX];
 	pid_t sec_mod_pid;
+
+	struct sockaddr_un secmod_addr;
+	unsigned secmod_addr_len;
 	
 	unsigned active_clients;
+	time_t start_time;
 
 	void * auth_extra;
+	struct config_mod_st *config_module;
 
-	struct ctl_list_st ctl_list;
+#ifdef HAVE_DBUS
 	void * ctl_ctx;
+#else
+	int ctl_fd;
+#endif
+	void *main_pool; /* talloc main pool */
 } main_server_st;
 
 void clear_lists(main_server_st *s);
@@ -206,7 +223,6 @@ int user_connected(main_server_st *s, struct proc_st* cur);
 void user_disconnected(main_server_st *s, struct proc_st* cur);
 
 void expire_tls_sessions(main_server_st *s);
-void expire_zombies(main_server_st* s);
 
 int send_udp_fd(main_server_st* s, struct proc_st * proc, int fd);
 
@@ -220,6 +236,9 @@ int handle_resume_fetch_req(main_server_st* s, struct proc_st * proc,
 int handle_resume_store_req(main_server_st* s, struct proc_st *proc,
   			   const SessionResumeStoreReqMsg *);
 
+int session_open(main_server_st * s, struct proc_st *proc);
+int session_close(main_server_st * s, struct proc_st *proc);
+
 void 
 __attribute__ ((format(printf, 4, 5)))
     _mslog(const main_server_st * s, const struct proc_st* proc,
@@ -237,36 +256,23 @@ void  mslog_hex(const main_server_st * s, const struct proc_st* proc,
     	int priority, const char *prefix, uint8_t* bin, unsigned bin_size, unsigned b64);
 
 int open_tun(main_server_st* s, struct proc_st* proc);
+void close_tun(main_server_st* s, struct proc_st* proc);
 int set_tun_mtu(main_server_st* s, struct proc_st * proc, unsigned mtu);
 
-int send_auth_reply_msg(main_server_st* s, struct proc_st* proc);
-
-int send_auth_reply(main_server_st* s, struct proc_st* proc,
-			AuthReplyMsg__AUTHREP r);
+int send_cookie_auth_reply(main_server_st* s, struct proc_st* proc,
+			AUTHREP r);
 
 int handle_auth_cookie_req(main_server_st* s, struct proc_st* proc,
  			   const AuthCookieRequestMsg * req);
-int generate_cookie(main_server_st *s, struct proc_st* proc);
-int handle_auth_init(main_server_st *s, struct proc_st* proc,
-		     const AuthInitMsg * req);
-int handle_auth_reinit(main_server_st *s, struct proc_st** proc,
-		     const AuthReinitMsg * req);
-int handle_auth_req(main_server_st *s, struct proc_st* proc,
-		     const AuthRequestMsg * req);
 
 int check_multiple_users(main_server_st *s, struct proc_st* proc);
-
-void add_to_ip_ban_list(main_server_st* s, struct sockaddr_storage *addr, socklen_t addr_len);
-void expire_banned(main_server_st* s);
-int check_if_banned(main_server_st* s, struct sockaddr_storage *addr, socklen_t addr_len);
-
 int handle_script_exit(main_server_st *s, struct proc_st* proc, int code);
 
 void run_sec_mod(main_server_st * s);
 
-int parse_group_cfg_file(main_server_st* s, const char* file, struct group_cfg_st *config);
-
-void del_additional_config(struct group_cfg_st* config);
+struct proc_st *new_proc(main_server_st * s, pid_t pid, int cmd_fd,
+			struct sockaddr_storage *remote_addr, socklen_t remote_addr_len,
+			uint8_t *sid, size_t sid_size);
 void remove_proc(main_server_st* s, struct proc_st *proc, unsigned k);
 void proc_to_zombie(main_server_st* s, struct proc_st *proc);
 
@@ -277,7 +283,7 @@ int send_msg_to_worker(main_server_st* s, struct proc_st* proc, uint8_t cmd,
 	    const void* msg, pack_size_func get_size, pack_func pack)
 {
 	mslog(s, proc, LOG_DEBUG, "sending message '%s' to worker", cmd_request_to_str(cmd));
-	return send_msg(proc->fd, cmd, msg, get_size, pack);
+	return send_msg(proc, proc->fd, cmd, msg, get_size, pack);
 }
 
 inline static
@@ -285,12 +291,9 @@ int send_socket_msg_to_worker(main_server_st* s, struct proc_st* proc, uint8_t c
 		int socketfd, const void* msg, pack_size_func get_size, pack_func pack)
 {
 	mslog(s, proc, LOG_DEBUG, "sending (socket) message %u to worker", (unsigned)cmd);
-	return send_socket_msg(proc->fd, cmd, socketfd, msg, get_size, pack);
+	return send_socket_msg(proc, proc->fd, cmd, socketfd, msg, get_size, pack);
 }
 
-void ctl_handle_commands(main_server_st* s, struct ctl_handler_st* ctl);
-int ctl_handler_init(main_server_st* s);
-void ctl_handler_deinit(main_server_st* s);
 void request_reload(int signo);
 void request_stop(int signo);
 
