@@ -8,7 +8,7 @@
  * the Free Software Foundation, either version 2 of the License, or
  * (at your option) any later version.
  *
- * GnuTLS is distributed in the hope that it will be useful, but
+ * ocserv is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
@@ -34,7 +34,9 @@
 #include <getpass.h>
 #include <minmax.h>
 
-static const char alphabet[] = 
+#define DEFAULT_OCPASSWD "/etc/ocserv/ocpasswd"
+
+static const char alphabet[] =
 	"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ./";
 
 #define SALT_SIZE 16
@@ -64,9 +66,8 @@ crypt_int(const char *fpasswd, const char *username, const char *groupname,
 		exit(1);
 	}
 
-	memset(salt, 0, sizeof(salt));
-	p = salt;
-	p += snprintf(salt, sizeof(salt), "$5$");
+	strcpy(salt, "$5$");
+	p = salt + 3;
 
 	for (i = 0; i < sizeof(_salt); i++) {
 		*p = alphabet[_salt[i] % (sizeof(alphabet) - 1)];
@@ -83,7 +84,7 @@ crypt_int(const char *fpasswd, const char *username, const char *groupname,
 		cr_passwd = crypt(passwd, salt);
 	}
 	if (cr_passwd == NULL) {
-		fprintf(stderr, "Error in crypt()\n");
+		fprintf(stderr, "Error in crypt().\n");
 		exit(1);
 	}
 
@@ -96,13 +97,13 @@ crypt_int(const char *fpasswd, const char *username, const char *groupname,
 
 	snprintf(tmp_passwd, tmp_passwd_len, "%s.tmp", fpasswd);
 	if (stat(tmp_passwd, &st) != -1) {
-		fprintf(stderr, "file '%s' is locked\n", fpasswd);
+		fprintf(stderr, "file '%s' is locked.\n", fpasswd);
 		exit(1);
 	}
 
 	fd2 = fopen(tmp_passwd, "w");
 	if (fd2 == NULL) {
-		fprintf(stderr, "Cannot open '%s' for write\n", tmp_passwd);
+		fprintf(stderr, "Cannot open '%s' for writing.\n", tmp_passwd);
 		exit(1);
 	}
 
@@ -126,14 +127,79 @@ crypt_int(const char *fpasswd, const char *username, const char *groupname,
 		}
 		free(line);
 		fclose(fd);
-		
+
 		if (found == 0)
 			fprintf(fd2, "%s:%s:%s\n", username, groupname, cr_passwd);
 	}
 
 	fclose(fd2);
 
-	rename(tmp_passwd, fpasswd);
+	ret = rename(tmp_passwd, fpasswd);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot write to '%s'.\n", fpasswd);
+		exit(1);
+	}
+	free(tmp_passwd);
+}
+
+static void
+delete_user(const char *fpasswd, const char *username)
+{
+	FILE * fd, *fd2;
+	char *tmp_passwd;
+	char *line, *p;
+	unsigned fpasswd_len = strlen(fpasswd);
+	unsigned tmp_passwd_len;
+	unsigned username_len = strlen(username);
+	int ret;
+	ssize_t len, l;
+	size_t line_size;
+	struct stat st;
+
+	tmp_passwd_len = fpasswd_len + 5;
+	tmp_passwd = malloc(tmp_passwd_len);
+
+	snprintf(tmp_passwd, tmp_passwd_len, "%s.tmp", fpasswd);
+	if (stat(tmp_passwd, &st) != -1) {
+		fprintf(stderr, "file '%s' is locked.\n", fpasswd);
+		exit(1);
+	}
+
+	fd = fopen(fpasswd, "r");
+	if (fd == NULL) {
+		fprintf(stderr, "Cannot open '%s' for reading.\n", fpasswd);
+		exit(1);
+	}
+
+	fd2 = fopen(tmp_passwd, "w");
+	if (fd2 == NULL) {
+		fprintf(stderr, "Cannot open '%s' for writing.\n", tmp_passwd);
+		exit(1);
+	}
+
+	line = NULL;
+	while ((len = getline(&line, &line_size, fd)) > 0) {
+		p = strchr(line, ':');
+		if (p == NULL)
+			continue;
+
+		l = p-line;
+		if (l == username_len && strncmp(line, username, l) == 0) {
+			continue;
+		} else {
+			fwrite(line, 1, len, fd2);
+		}
+	}
+
+	free(line);
+	fclose(fd);
+	fclose(fd2);
+
+	ret = rename(tmp_passwd, fpasswd);
+	if (ret == -1) {
+		fprintf(stderr, "Cannot write to '%s'.\n", fpasswd);
+		exit(1);
+	}
 	free(tmp_passwd);
 }
 
@@ -146,6 +212,7 @@ lock_user(const char *fpasswd, const char *username)
 	unsigned fpasswd_len = strlen(fpasswd);
 	unsigned tmp_passwd_len;
 	unsigned username_len = strlen(username);
+	int ret;
 	ssize_t len, l;
 	size_t line_size;
 	struct stat st;
@@ -155,19 +222,19 @@ lock_user(const char *fpasswd, const char *username)
 
 	snprintf(tmp_passwd, tmp_passwd_len, "%s.tmp", fpasswd);
 	if (stat(tmp_passwd, &st) != -1) {
-		fprintf(stderr, "file '%s' is locked\n", fpasswd);
+		fprintf(stderr, "file '%s' is locked.\n", fpasswd);
 		exit(1);
 	}
 
 	fd = fopen(fpasswd, "r");
 	if (fd == NULL) {
-		fprintf(stderr, "Cannot open '%s' for read\n", fpasswd);
+		fprintf(stderr, "Cannot open '%s' for reading.\n", fpasswd);
 		exit(1);
 	}
 
 	fd2 = fopen(tmp_passwd, "w");
 	if (fd2 == NULL) {
-		fprintf(stderr, "Cannot open '%s' for write\n", tmp_passwd);
+		fprintf(stderr, "Cannot open '%s' for writing.\n", tmp_passwd);
 		exit(1);
 	}
 
@@ -183,7 +250,7 @@ lock_user(const char *fpasswd, const char *username)
 			if (p == NULL)
 				continue;
 			p++;
-			
+
 			l = p-line;
 			fwrite(line, 1, l, fd2);
 			fputc('!', fd2);
@@ -197,7 +264,11 @@ lock_user(const char *fpasswd, const char *username)
 	fclose(fd);
 	fclose(fd2);
 
-	rename(tmp_passwd, fpasswd);
+	ret = rename(tmp_passwd, fpasswd);
+	if (ret == -1) {
+		fprintf(stderr, "Cannot write to '%s'.\n", fpasswd);
+		exit(1);
+	}
 	free(tmp_passwd);
 }
 
@@ -210,6 +281,7 @@ unlock_user(const char *fpasswd, const char *username)
 	unsigned fpasswd_len = strlen(fpasswd);
 	unsigned tmp_passwd_len;
 	unsigned username_len = strlen(username);
+	int ret;
 	ssize_t len, l;
 	size_t line_size;
 	struct stat st;
@@ -219,19 +291,19 @@ unlock_user(const char *fpasswd, const char *username)
 
 	snprintf(tmp_passwd, tmp_passwd_len, "%s.tmp", fpasswd);
 	if (stat(tmp_passwd, &st) != -1) {
-		fprintf(stderr, "file '%s' is locked\n", fpasswd);
+		fprintf(stderr, "file '%s' is locked.\n", fpasswd);
 		exit(1);
 	}
 
 	fd = fopen(fpasswd, "r");
 	if (fd == NULL) {
-		fprintf(stderr, "Cannot open '%s' for read\n", fpasswd);
+		fprintf(stderr, "Cannot open '%s' for reading.\n", fpasswd);
 		exit(1);
 	}
 
 	fd2 = fopen(tmp_passwd, "w");
 	if (fd2 == NULL) {
-		fprintf(stderr, "Cannot open '%s' for write\n", tmp_passwd);
+		fprintf(stderr, "Cannot open '%s' for writing.\n", tmp_passwd);
 		exit(1);
 	}
 
@@ -247,7 +319,7 @@ unlock_user(const char *fpasswd, const char *username)
 			if (p == NULL)
 				continue;
 			p++;
-			
+
 			l = p-line;
 			fwrite(line, 1, l, fd2);
 
@@ -263,7 +335,11 @@ unlock_user(const char *fpasswd, const char *username)
 	fclose(fd);
 	fclose(fd2);
 
-	rename(tmp_passwd, fpasswd);
+	ret = rename(tmp_passwd, fpasswd);
+	if (ret == -1) {
+		fprintf(stderr, "Cannot write to '%s'.\n", fpasswd);
+		exit(1);
+	}
 	free(tmp_passwd);
 }
 
@@ -271,7 +347,8 @@ int main(int argc, char **argv)
 {
 	int ret, optct;
 	const char *username, *groupname, *fpasswd;
-	const char* passwd;
+	char* passwd;
+	size_t l, i;
 
 	if ((ret = gnutls_global_init()) < 0) {
 		fprintf(stderr, "global_init: %s\n", gnutls_strerror(ret));
@@ -283,19 +360,18 @@ int main(int argc, char **argv)
 	optct = optionProcess(&ocpasswdOptions, argc, argv);
 	argc -= optct;
 	argv += optct;
-	
+
 	if (argc > 0)
 		username = argv[0];
 	else {
-		fprintf(stderr, "Please specify a user\n");
-		exit(1); 
+		optionUsage(&ocpasswdOptions, 1);
+		exit(1);
 	}
 
-	if (HAVE_OPT(PASSWD))
+	if (HAVE_OPT(PASSWD)) {
 		fpasswd = OPT_ARG(PASSWD);
-	else {
-		fprintf(stderr, "Password file was not specified\n");
-		exit(1);
+	} else {
+		fpasswd = DEFAULT_OCPASSWD;
 	}
 
 	if (HAVE_OPT(GROUPNAME))
@@ -303,35 +379,52 @@ int main(int argc, char **argv)
 	else {
 		groupname = "*";
 	}
-	
+
 	if (HAVE_OPT(LOCK))
 		lock_user(fpasswd, username);
 	else if (HAVE_OPT(UNLOCK))
 		unlock_user(fpasswd, username);
+	else if (HAVE_OPT(DELETE))
+		delete_user(fpasswd, username);
 	else { /* set password */
-		char* p2;
 
-		passwd = getpass("Enter password: ");
-		if (passwd == NULL) {
-			fprintf(stderr, "Please specify a password\n");
-			exit(1); 
-		}
-		p2 = strdup(passwd);
-		passwd = getpass("Re-enter password: ");
-		if (passwd == NULL) {
-			fprintf(stderr, "Please specify a password\n");
-			exit(1); 
-		}
+		if (isatty(STDIN_FILENO)) {
+			char* p2;
 
-		if (p2 == NULL || strcmp(passwd, p2) != 0) {
-			fprintf(stderr, "Passwords do not match\n");
-			exit(1); 
+			passwd = getpass("Enter password: ");
+			if (passwd == NULL) {
+				fprintf(stderr, "Please specify a password\n");
+				exit(1);
+			}
+
+
+			p2 = strdup(passwd);
+			passwd = getpass("Re-enter password: ");
+			if (passwd == NULL) {
+				fprintf(stderr, "Please specify a password\n");
+				exit(1);
+			}
+
+			if (p2 == NULL || strcmp(passwd, p2) != 0) {
+				fprintf(stderr, "Passwords do not match\n");
+				exit(1);
+			}
+			free(p2);
+		} else {
+			passwd = NULL;
+			l = getline(&passwd, &i, stdin);
+			if (l <= 1) {
+				fprintf(stderr, "Please specify a password\n");
+				exit(1);
+			}
+
+			if (passwd[l-1] == '\n')
+				passwd[l-1] = 0;
 		}
-		free(p2);
 
 		crypt_int(fpasswd, username, groupname, passwd);
 	}
-	
+
 	gnutls_global_deinit();
 	return 0;
 }
