@@ -44,6 +44,18 @@
 # define _ATTR_PACKED
 #endif
 
+typedef enum {
+	SOCK_TYPE_TCP,
+	SOCK_TYPE_UDP,
+	SOCK_TYPE_UNIX
+} sock_type_t;
+
+#define DEBUG_BASIC 1
+#define DEBUG_HTTP  2
+#define DEBUG_TRANSFERRED 5
+#define DEBUG_TLS   9
+
+#define DEFAULT_DPD_TIME 600
 
 #define AC_PKT_DATA             0	/* Uncompressed data */
 #define AC_PKT_DPD_OUT          3	/* Dead Peer Detection */
@@ -53,13 +65,17 @@
 #define AC_PKT_COMPRESSED       8	/* Compressed data */
 #define AC_PKT_TERM_SERVER      9	/* Server kick */
 
+#define REKEY_METHOD_SSL 1
+#define REKEY_METHOD_NEW_TUNNEL 2
+
 extern int syslog_open;
 
 /* the first is generic, for the methods that require a username password */
 #define AUTH_TYPE_USERNAME_PASS (1<<0)
 #define AUTH_TYPE_PAM (1<<1 | AUTH_TYPE_USERNAME_PASS)
-#define AUTH_TYPE_CERTIFICATE (1<<2)
-#define AUTH_TYPE_PLAIN (1<<3 | AUTH_TYPE_USERNAME_PASS)
+#define AUTH_TYPE_PLAIN (1<<2 | AUTH_TYPE_USERNAME_PASS)
+#define AUTH_TYPE_CERTIFICATE (1<<3)
+#define AUTH_TYPE_CERTIFICATE_OPT (1<<4|AUTH_TYPE_CERTIFICATE)
 
 #define ERR_SUCCESS 0
 #define ERR_BAD_COMMAND -2
@@ -71,24 +87,24 @@ extern int syslog_open;
 #define ERR_NO_IP -8
 #define ERR_PARSING -9
 #define ERR_EXEC -10
-#define ERR_WORKER_TERMINATED -11
+#define ERR_PEER_TERMINATED -11
 #define ERR_CTL -12
+#define ERR_NO_CMD_FD -13
+
+#define ERR_WORKER_TERMINATED ERR_PEER_TERMINATED
 
 #define LOG_HTTP_DEBUG 2048
+#define LOG_TRANSFER_DEBUG 2049
 
+#define MAX_AUTH_SECS 40
 #define MAX_CIPHERSUITE_NAME 64
 #define MAX_DTLS_CIPHERSUITE_NAME 24
 #define MAX_MSG_SIZE 256
 #define SID_SIZE 12
 
-#define MAX_ZOMBIE_SECS 240
-
 typedef enum {
-	AUTH_INIT = 1,
-	AUTH_REP = 2,
-	AUTH_REQ = 3,
+	AUTH_COOKIE_REP = 2,
 	AUTH_COOKIE_REQ = 4,
-	AUTH_MSG = 5,
 	RESUME_STORE_REQ = 6,
 	RESUME_DELETE_REQ = 7,
 	RESUME_FETCH_REQ = 8,
@@ -97,14 +113,19 @@ typedef enum {
 	CMD_TUN_MTU = 11,
 	CMD_TERMINATE = 12,
 	CMD_SESSION_INFO = 13,
-	AUTH_REINIT = 14,
+	CMD_CLI_STATS = 15,
+
+	SM_CMD_AUTH_INIT = 120,
+	SM_CMD_AUTH_CONT,
+	SM_CMD_AUTH_REP,
+	SM_CMD_DECRYPT,
+	SM_CMD_SIGN,
+	SM_CMD_AUTH_SESSION_OPEN,
+	SM_CMD_AUTH_SESSION_CLOSE,
+	SM_CMD_AUTH_SESSION_REPLY,
 } cmd_request_t;
 
-typedef struct 
-{
-	struct htable ht;
-	unsigned int entries;
-} hash_db_st;
+#define MAX_IP_STR 46
 
 struct group_cfg_st {
 	/* routes to be forwarded to the client */
@@ -115,44 +136,61 @@ struct group_cfg_st {
 	char **iroutes;
 	unsigned int iroutes_size;
 
-	char *ipv4_dns;
-	char *ipv6_dns;
-	char *ipv4_nbns;
-	char *ipv6_nbns;
+	char **dns;
+	unsigned int dns_size;
+
+	char **nbns;
+	unsigned int nbns_size;
+
 	char *ipv4_network;
 	char *ipv6_network;
+	unsigned ipv6_prefix;
 	char *ipv4_netmask;
 	char *ipv6_netmask;
 	
 	char *cgroup;
 
+	char *xml_config_file;
+
 	size_t rx_per_sec;
 	size_t tx_per_sec;
-	
+
+	unsigned deny_roaming; /* whether the user is allowed to re-use cookies from another IP */
 	unsigned net_priority;
+	unsigned no_udp; /* whether to disable UDP for this user */
+	unsigned require_cert; /* when optional certificate auth is selected require a certificate */
 };
 
 struct vpn_st {
-	char *name;	/* device name */
+	char name[IFNAMSIZ];
 	char *ipv4_netmask;
+	char *ipv4_network;
 	char *ipv4;
 	char *ipv4_local; /* local IPv4 address */
 	char *ipv6_netmask;
+	char *ipv6_network;
+	unsigned ipv6_prefix;
+
 	char *ipv6;
 	char *ipv6_local; /* local IPv6 address */
-	char *ipv4_dns;
-	char *ipv6_dns;
-	char *ipv4_nbns;
-	char *ipv6_nbns;
 	unsigned int mtu;
+
 	char **routes;
 	unsigned int routes_size;
+
+	char **dns;
+	unsigned int dns_size;
+
+	char **nbns;
+	unsigned int nbns_size;
 };
 
 struct cfg_st {
 	char *name; /* server name */
 	unsigned int port;
 	unsigned int udp_port;
+	char* unix_conn_file;
+
 	char *pin_file;
 	char *srk_pin_file;
 	char **cert;
@@ -166,41 +204,69 @@ struct cfg_st {
 	char *cert_user_oid;	/* The OID that will be used to extract the username */
 	char *cert_group_oid;	/* The OID that will be used to extract the groupname */
 	unsigned int auth_types;	/* or'ed sequence of AUTH_TYPE */
-	char *plain_passwd;	/* the password file in plain auth mode */
+	unsigned session_control; /* whether to use the session control part of authentication (PAM) */
+	char *auth_additional;	/* the additional string specified in the auth methode */
 	gnutls_certificate_request_t cert_req;
 	char *priorities;
 	char *chroot_dir;	/* where the xml files are served from */
 	char *banner;
 	char *ocsp_response; /* file with the OCSP response */
 	char *default_domain; /* domain to be advertised */
-	
+
+	char **group_list; /* select_group */
+	unsigned int group_list_size;
+
+	char **friendly_group_list; /* the same size as group_list_size */
+
+	char *default_select_group;
+
+	char **custom_header;
+	unsigned custom_header_size;;
+
+	char **split_dns;
+	unsigned split_dns_size;;
+
 	char* socket_file_prefix;
-	time_t cookie_validity;	/* in seconds */
+
+	unsigned deny_roaming; /* whether a cookie is restricted to a single IP */
+	time_t cookie_timeout;	/* in seconds */
+
+	time_t rekey_time;	/* in seconds */
+	unsigned rekey_method; /* REKEY_METHOD_ */
+
 	time_t min_reauth_time;	/* after a failed auth, how soon one can reauthenticate -> in seconds */
+
+	unsigned seccomp; /* whether seccomp should be enabled or not */
+
 	unsigned auth_timeout; /* timeout of HTTP auth */
+	unsigned idle_timeout; /* timeout when idle */
+	unsigned mobile_idle_timeout; /* timeout when a mobile is idle */
 	unsigned keepalive;
 	unsigned dpd;
+	unsigned mobile_dpd;
 	unsigned foreground;
-	unsigned tls_debug;
-	unsigned http_debug;
 	unsigned debug;
 	unsigned max_clients;
 	unsigned max_same_clients;
 	unsigned use_utmp;
 	unsigned use_dbus; /* whether the D-BUS service is registered */
+	unsigned use_occtl; /* whether support for the occtl tool will be enabled */
+	char* occtl_socket_file;
+
 	unsigned try_mtu; /* MTU discovery enabled */
 	unsigned cisco_client_compat; /* do not require client certificate, 
 	                               * and allow auth to complete in different
 	                               * TCP sessions. */
 	unsigned rate_limit_ms; /* if non zero force a connection every rate_limit milliseconds */
 	unsigned ping_leases; /* non zero if we need to ping prior to leasing */
-	
+
 	size_t rx_per_sec;
 	size_t tx_per_sec;
 	unsigned net_priority;
 
 	unsigned output_buffer;
 	unsigned default_mtu;
+	unsigned predictable_ips; /* boolean */
 
 	char *route_add_cmd;
 	char *route_del_cmd;
@@ -209,6 +275,7 @@ struct cfg_st {
 	char *disconnect_script;
 	
 	char *cgroup;
+	char *proxy_url;
 
 #ifdef ANYCONNECT_CLIENT_COMPAT
 	char *xml_config_file;
@@ -222,6 +289,8 @@ struct cfg_st {
 	/* additional configuration files */
 	char *per_group_dir;
 	char *per_user_dir;
+	char *default_group_conf;
+	char *default_user_conf;
 	
 	/* the tun network */
 	struct vpn_st network;
@@ -233,7 +302,7 @@ struct main_server_st;
 
 #define MAX_BANNER_SIZE 256
 #define MAX_USERNAME_SIZE 64
-#define MAX_AGENT_NAME 32
+#define MAX_AGENT_NAME 48
 #define MAX_PASSWORD_SIZE 64
 #define TLS_MASTER_SIZE 48
 #define MAX_HOSTNAME_SIZE MAX_USERNAME_SIZE
@@ -242,11 +311,9 @@ struct main_server_st;
 
 #define MAX_CONFIG_ENTRIES 64
 
-#define MAX_ROUTES 32
-
 #include <tun.h>
 
-const char *human_addr2(const struct sockaddr *sa, socklen_t salen,
+char *human_addr2(const struct sockaddr *sa, socklen_t salen,
 		       void *buf, size_t buflen, unsigned full);
 
 #define human_addr(x, y, z, w) human_addr2(x, y, z, w, 1)
@@ -261,6 +328,7 @@ const char *human_addr2(const struct sockaddr *sa, socklen_t salen,
 #define SA_IN6_PORT(p) (((struct sockaddr_in6 *)(p))->sin6_port)
 
 #define SA_IN_P_GENERIC(addr, size) ((size==sizeof(struct sockaddr_in))?SA_IN_U8_P(addr):SA_IN6_U8_P(addr))
+#define SA_IN_P_TYPE(addr, type) ((type==AF_INET)?SA_IN_U8_P(addr):SA_IN6_U8_P(addr))
 #define SA_IN_SIZE(size) ((size==sizeof(struct sockaddr_in))?sizeof(struct in_addr):sizeof(struct in6_addr))
 
 /* macros */
