@@ -32,6 +32,9 @@
 #include <tlslib.h>
 #include <common.h>
 #include <str.h>
+#include <worker-bandwidth.h>
+#include <sys/un.h>
+#include <sys/uio.h>
 
 typedef enum {
 	UP_DISABLED,
@@ -42,15 +45,27 @@ typedef enum {
 	UP_ACTIVE
 } udp_port_state_t;
 
+#define STR_HDR_COOKIE "Cookie"
+#define STR_HDR_USER_AGENT "User-Agent"
+#define STR_HDR_CONNECTION "Connection"
+#define STR_HDR_MS "X-DTLS-Master-Secret"
+#define STR_HDR_CS "X-DTLS-CipherSuite"
+#define STR_HDR_CMTU "X-CSTP-Base-MTU"
+#define STR_HDR_ATYPE "X-CSTP-Address-Type"
+#define STR_HDR_HOST "X-CSTP-Hostname"
+#define STR_HDR_FULL_IPV6 "X-CSTP-Full-IPv6-Capability"
+#define STR_HDR_DEVICE_TYPE "X-AnyConnect-Identifier-DeviceType"
+
 enum {
 	HEADER_COOKIE = 1,
 	HEADER_MASTER_SECRET,
 	HEADER_HOSTNAME,
-	HEADER_CSTP_MTU,
+	HEADER_CSTP_BASE_MTU,
 	HEADER_CSTP_ATYPE,
-	HEADER_DTLS_MTU,
+	HEADER_DEVICE_TYPE,
 	HEADER_DTLS_CIPHERSUITE,
 	HEADER_CONNECTION,
+	HEADER_FULL_IPV6,
 	HEADER_USER_AGENT,
 };
 
@@ -64,9 +79,16 @@ enum {
 	S_AUTH_INACTIVE = 0,
 	S_AUTH_INIT,
 	S_AUTH_REQ,
+	S_AUTH_COOKIE,
 	S_AUTH_COMPLETE
 };
 
+enum {
+	AGENT_UNKNOWN,
+	AGENT_OPENCONNECT_V3,
+	AGENT_OPENCONNECT
+};
+
 typedef struct dtls_ciphersuite_st {
 	const char* oc_name;
 	const char* gnutls_name; /* the gnutls priority string to set */
@@ -85,10 +107,11 @@ struct http_req_st {
 
 	char hostname[MAX_HOSTNAME_SIZE];
 	char user_agent[MAX_AGENT_NAME];
+	unsigned user_agent_type;;
+
 	unsigned int next_header;
-	unsigned char cookie[COOKIE_SIZE];
-	unsigned int cookie_set;
-	unsigned int sid_cookie_set;
+
+	unsigned int is_mobile;
 
 	unsigned char master_secret[TLS_MASTER_SIZE];
 	unsigned int master_secret_set;
@@ -100,8 +123,7 @@ struct http_req_st {
 
 	unsigned int headers_complete;
 	unsigned int message_complete;
-	unsigned dtls_mtu;
-	unsigned cstp_mtu;
+	unsigned base_mtu;
 	
 	unsigned no_ipv4;
 	unsigned no_ipv6;
@@ -112,16 +134,23 @@ typedef struct worker_st {
 	gnutls_session_t session;
 	gnutls_session_t dtls_session;
 
+	struct http_req_st req;
+
 	/* inique session identifier */
 	uint8_t sid[SID_SIZE];
+	unsigned int sid_set;
 
 	int cmd_fd;
 	int conn_fd;
+	sock_type_t conn_type; /* AF_UNIX or something else */
 	
 	http_parser *parser;
 	struct cfg_st *config;
 	unsigned int auth_state; /* S_AUTH */
 
+	struct sockaddr_un secmod_addr;	/* sec-mod unix address */
+	socklen_t secmod_addr_len;
+
 	struct sockaddr_storage remote_addr;	/* peer's address */
 	socklen_t remote_addr_len;
 	int proto; /* AF_INET or AF_INET6 */
@@ -129,39 +158,80 @@ typedef struct worker_st {
 	/* for dead peer detection */
 	time_t last_msg_udp;
 	time_t last_msg_tcp;
+
+	time_t last_nc_msg; /* last message that wasn't control, on any channel */
+
 	time_t last_periodic_check;
 
 	/* set after authentication */
 	int udp_fd;
 	udp_port_state_t udp_state;
+	time_t udp_recv_time; /* time last udp packet was received */
 
 	/* protection from multiple rehandshakes */
 	time_t last_tls_rehandshake;
+	time_t last_dtls_rehandshake;
 
 	/* for mtu trials */
 	unsigned last_good_mtu;
 	unsigned last_bad_mtu;
+
+	/* bandwidth stats */
+	bandwidth_st b_tx;
+	bandwidth_st b_rx;
+
+	/* ws->conn_mtu: The MTU of the plaintext data we can send to the client.
+	 *  It also matches the MTU of the TUN device. Note that this is
+	 *  the same as the 'real' MTU of the connection, minus the IP+UDP+CSTP headers
+	 *  and the DTLS crypto overhead. */
 	unsigned conn_mtu;
+	unsigned crypto_overhead; /* estimated overhead of DTLS ciphersuite + DTLS CSTP HEADER */
 	
+	/* Indicates whether the new IPv6 headers will
+	 * be sent or the old */
+	unsigned full_ipv6;
+
 	/* Buffer used by worker */
-	uint8_t * buffer;
+	uint8_t buffer[16*1024];
 	unsigned buffer_size;
 
 	/* the following are set only if authentication is complete */
-	char tun_name[IFNAMSIZ];
+
 	char username[MAX_USERNAME_SIZE];
+	char groupname[MAX_GROUPNAME_SIZE];
+
+	char cert_username[MAX_USERNAME_SIZE];
+	char **cert_groups;
+	unsigned cert_groups_size;
+
 	char hostname[MAX_HOSTNAME_SIZE];
-	uint8_t cookie[COOKIE_SIZE];
+	uint8_t *cookie;
+	unsigned cookie_size;
+
+	unsigned int cookie_set;
+
 	uint8_t master_secret[TLS_MASTER_SIZE];
 	uint8_t session_id[GNUTLS_MAX_SESSION_ID];
 	unsigned cert_auth_ok;
 	int tun_fd;
+
+	/* tun device stats */
+	uint64_t tun_bytes_in;
+	uint64_t tun_bytes_out;
+
+	/* information on the tun device addresses and network */
+	struct vpn_st vinfo;
+	unsigned default_route;
 	
 	/* additional data - received per user or per group */
 	unsigned routes_size;
-	char* routes[MAX_ROUTES];
-	
-	struct http_req_st req;
+	char** routes;
+	unsigned dns_size;
+	char** dns;
+	unsigned nbns_size;
+	char** nbns;
+
+	void *main_pool; /* to be used only on deinitialization */
 } worker_st;
 
 void vpn_server(struct worker_st* ws);
@@ -176,6 +246,7 @@ int get_empty_handler(worker_st *server, unsigned http_ver);
 int get_config_handler(worker_st *ws, unsigned http_ver);
 int get_string_handler(worker_st *ws, unsigned http_ver);
 int get_dl_handler(worker_st *ws, unsigned http_ver);
+int get_cert_names(worker_st * ws, const gnutls_datum_t * raw);
 
 void set_resume_db_funcs(gnutls_session_t);
 
@@ -191,19 +262,21 @@ void __attribute__ ((format(printf, 3, 4)))
 # define oclog _oclog
 #endif
 
-int get_rt_vpn_info(worker_st * ws,
-                    struct vpn_st* vinfo, char* buffer, size_t buffer_size);
+int complete_vpn_info(worker_st * ws,
+                    struct vpn_st* vinfo);
+unsigned check_if_default_route(char **routes, unsigned routes_size);
 
 int send_tun_mtu(worker_st *ws, unsigned int mtu);
 int handle_worker_commands(struct worker_st *ws);
 int disable_system_calls(struct worker_st *ws);
+void ocsigaltstack(struct worker_st *ws);
 
 inline static
 int send_msg_to_main(worker_st *ws, uint8_t cmd, 
 	    const void* msg, pack_size_func get_size, pack_func pack)
 {
 	oclog(ws, LOG_DEBUG, "sending message '%s' to main", cmd_request_to_str(cmd));
-	return send_msg(ws->cmd_fd, cmd, msg, get_size, pack);
+	return send_msg(ws, ws->cmd_fd, cmd, msg, get_size, pack);
 }
 
 #endif
